//---------------------------------------------------------------------------

#ifndef UtListaH
#define UtListaH

#include "UErrores.h"
#include <fstream>
using namespace std;
//---------------------------------------------------------------------------

template <class T> class Lista
{
public:
        Lista() {pCab = new nodo; pi = new puntoInt;
                 pCab->sig = pCab; pCab->ant = pCab;
                 pi->nod = pCab; pi->cont = cont = 0;};
        virtual ~Lista() throw() {clear(); delete pCab; delete pi;};
        virtual void clear() throw();
        virtual unsigned int size() const throw() {return cont;};
        virtual bool empty() const throw() {return cont == 0;};
        virtual T& at(int Pos) throw(ErrRango);
        virtual void at(int Pos, T& E) const throw(ErrRango) {E=at(Pos);};
        virtual const T& at(int Pos) const throw(ErrRango) {return ((Lista<T>*)this)->at(Pos);};
        virtual void insert(int Pos, const T& E) throw(ErrRango, bad_alloc);
        virtual void erase(int Pos) throw(ErrRango);
        void front(T& E) const throw(ErrVacia);
        T& front() throw(ErrVacia);
        void push_front(const T& E) throw(bad_alloc);
        void pop_front() throw(ErrVacia);
        void back(T& E) const throw(ErrVacia);
        T& back() const throw(ErrVacia);
        void push_back(const T& E) throw(bad_alloc);
        void pop_back() throw(ErrVacia);
        bool operator==(const Lista<T>& list) const;
        bool operator!=(const Lista<T>& list) const {return!(*this == list);};

protected:
        struct nodo
        {
                T elem;
                nodo *sig, *ant;
        };
        struct puntoInt
        {
                unsigned int cont;
                nodo *nod;
        };
        void iniciaPI() const throw(); //posiciona pi en la cabecera.
        void avanzaPI() const; //mueve pi al siguiente nodo.
        void retrocedePI() const; //mueve pi al nodo anterior.
        void posicionPI(int Pos) const throw(ErrRango); //mueve pi al nodo indicado por la posicin Pos.
        void elemPI(T& E) const;
        T& elemPI() const; //devuelve el elemento del nodo apuntado por pi.
        void elemAntPI(T& E) const; //devuelve el elemento del nodo anterior apuntado por pi.
        void insertPI(const T& E) throw(bad_alloc); //aade un nuevo nodo en la posicin indicada por pi (i.e. anterior a pCur).
        void quitaPI();
        unsigned int contPI() const {return pi->cont;};

private:
        nodo* pCab;
        puntoInt* pi;
        unsigned int cont;
        bool inRange(unsigned int pos) const {return pos <= cont;};
};
//---------------------------------------------------------------------------

template <class T>
void Lista<T>::iniciaPI() const throw()  //posiciona pi en la cabecera.
{
        pi->cont = 0;
        pi->nod = pCab->sig;
}

template <class T>
void Lista<T>::avanzaPI() const //mueve pi al siguiente nodo.
{
        if(inRange(pi->cont + 1))
        {
                pi->nod = pi->nod->sig;
                pi->cont++;
        }
        else
        {
                pi->nod = pCab->sig;
                pi->cont = 0;
        }
}

template <class T>
void Lista<T>::retrocedePI() const//mueve pi al nodo anterior.
{
        if(inRange(pi->cont) && (pi->cont > 0))
        {
                pi->nod = pi->nod->ant;
                pi->cont--;
        }
        else
        {
                pi->nod = pCab;
                pi->cont = cont;
        }
}

template <class T>
void Lista<T>::posicionPI(int Pos) const throw(ErrRango) //mueve pi al nodo indicado por la posicin Pos.
{
        if(inRange(Pos))
        {
                if(pi->cont > (unsigned int) Pos)
                {
                        while(pi->cont != (unsigned int) Pos)
                        {
                                retrocedePI();
                        }
                }
                else
                {
                        while(pi->cont != (unsigned int) Pos)
                        {
                                avanzaPI();
                        }
                }

        }
        else throw ErrRango();
}

template <class T>
void Lista<T>::elemPI(T& E) const
{
        E = pi->nodo->elem;
}

template <class T>
T& Lista<T>::elemPI() const  //devuelve el elemento del nodo apuntado por pi.
{
        return pi->nod->elem;
}

template <class T>
void Lista<T>::elemAntPI(T& E) const //devuelve el elemento del nodo anterior apuntado por pi.
{
        E = pi->nod->ant->elem;
}

template <class T>
void Lista<T>::insertPI(const T& E) throw(bad_alloc) //aade un nuevo nodo en la posicin indicada por pi (i.e. anterior a pCur).
{
        try
        {
                nodo* aux = new nodo;
                aux->sig = pi->nod;
                aux->ant = pi->nod->ant;
                aux->elem = E;
                pi->nod->ant->sig = aux;
                pi->nod->ant = aux;
                pi->nod = aux;
                ++cont;
        }
        catch (bad_alloc) {}
}

template <class T>
void Lista<T>::quitaPI()
{
        nodo* aux = pi->nod;
        pi->nod->ant->sig = aux->sig;
        pi->nod->sig->ant = aux->ant;
        pi->nod = aux->sig;
        delete aux;
        --cont;
}

template <class T>
void Lista<T>::insert(int Pos, const T& E) throw(ErrRango, bad_alloc)
{
        if(inRange(Pos))
        {
                posicionPI(Pos);
                insertPI(E);
        }
}

template <class T>
void Lista<T>::erase(int Pos) throw(ErrRango)
{
        if(inRange(Pos + 1))
        {
                posicionPI(Pos);
                quitaPI();
        }
}

template <class T>
void Lista<T>::clear() throw()
{
        if(cont > 0)
        {
                iniciaPI();
                while (pi->nod != pCab)
                {
                        quitaPI();
                }
                pCab->sig = pCab;
                pCab->ant = pCab;
                cont = 0;
        }
}

template <class T>
T& Lista<T>::at(int Pos) throw(ErrRango) //Invoca a posicionPI(Pos) y elemPI().
{
        if(inRange(Pos))
        {
                posicionPI(Pos);
                return pi->nod->elem;
        }
        else throw ErrRango("La casilla no existe.");
}

template <class T>
void Lista<T>::front(T& E) const throw(ErrVacia)
{
        if(cont > 0)
        {
                E = pCab->sig->elem;
        }
        else throw ErrVacia();
}

template <class T>
T& Lista<T>::front() throw(ErrVacia) //No utiliza el punto de inters.
{
        if(cont > 0)
        {
                return pCab->sig->elem;
        }
        else throw ErrVacia();
}

template <class T>
void Lista<T>::push_front(const T& E) throw(bad_alloc) //No modifica el punto de inters, pero no puede dejarlo incoherente.
{
        int aux = pi->cont;
        iniciaPI();
        insertPI(E);
        posicionPI(aux);
}

template <class T>
void Lista<T>::pop_front() throw(ErrVacia) //No utiliza el punto de inters, pero no puede dejarlo incoherente.
{
        int aux = pi->cont;
        iniciaPI();
        quitaPI();
        posicionPI(aux);
}

template <class T>
void Lista<T>::back(T& E) const throw(ErrVacia)
{
        if(cont > 0)
        {
                E = pCab->ant->elem;
        }
        else throw ErrVacia();
}

template <class T>
T& Lista<T>::back() const throw(ErrVacia) //No utiliza el punto de inters.
{
        if(cont > 0)
        {
                return pCab->ant->elem;
        }
        else throw ErrVacia();
}

template <class T>
void Lista<T>::push_back(const T& E) throw(bad_alloc) //No utiliza el punto de inters, pero no puede dejarlo incoherente.
{
        unsigned int aux = pi->cont;
        iniciaPI();
        retrocedePI();
        insertPI(E);
        posicionPI(aux);
}

template <class T>
void Lista<T>::pop_back() throw(ErrVacia) //No utiliza el punto de inters, pero no puede dejarlo incoherente.
{
        int aux = pi->cont;
        iniciaPI();
        retrocedePI();
        quitaPI();
        posicionPI(aux);
}

/*template <class T>
bool Lista<T>::salvar(ofstream& sal) const
{
        unsigned int aux = pi->cont;
        sal.write((char*)& cont, sizeof(unsigned int));
        sal.write((char*)& pi->cont, sizeof(unsigned int));
        if(cont > 0)
        {
                iniciaPI();
                while(pi->cont < cont)
                {
                        pi->nod->elem.salvar(sal);
                        avanzaPI();
                }
                posicionPI(aux);
        }
        return sal;
}

template <class T>
bool Lista<T>::cargar(ifstream& ent)
{
        clear();
        unsigned int listcont, picont;
        ent.read((char*)& listcont, sizeof(unsigned int));
        ent.read((char*)& picont, sizeof(unsigned int));
        if(listcont > 0)
        {
                T aux;
                while(listcont != cont)
                {
                        aux.cargar(ent);
                        insertPI(aux);
                        avanzaPI();
                }
                posicionPI(picont);
        }
        return ent;
} */

#endif
